/*
This software is published under the GNU GPL V3 license.
See http://www.gnu.org/licenses/gpl-3.0.en.html for Details.

Use Arduino IDE 1.5.6-r2 (or a newer one if it works ;-).

If you want to see text from the SD card then copy the file
"SD_test.txt" on the SD card (works with both cards).

After the SD and eeprom test the second LCD line shows the 
thermistor values in kOhm (9999 without thermistors).
*/

#include <LiquidCrystal.h>
#include <SPI.h>
#include <SD.h>
#include <Wire.h>    
#include "DueTimer.h"


// EEPROM: 
#define dev_adr 0x50    //Address of 24LC256 eeprom chip
unsigned int eep_adr = 5;  // Adresse im EEPROM
byte b1 = 0, b2 = 0;

// Zeiten Geschw. etc.
volatile unsigned long p1ms = 0, p1ms_start = 0, p1ms_fet = 0; // Programmlaufzeit in 0.1 ms:
volatile int lcd_aus = 0; // LCD-Ausgabe erfolgt nur ab und an
volatile float vv = 0.0; // Gescw. in (steps / 0.1ms)
volatile float tt = 0.0; // Zeit in (0.1ms)

// Drehencoder
volatile int enc_pin[3];
volatile int enc;

// SD-Karte:
int chipSelect = 10;  // für LCD-RADDS
byte sd_ok = 0;
byte sd_cd = 0;
int pos = 0;

// Stepper:
volatile int stp_soll[6], stp_ist[6], stp_0[6]; // Schritte der Motoren
volatile byte dir_pin[6]; // Stepper Dir pins: x, y, z, e0, e1, e2
volatile byte stp_pin[6]; // Stepper Step pins: x, y, z, e0, e1, e2
volatile byte ena_pin[6]; // Stepper enable pins: x, y, z, e0, e1, e2
volatile int motor; // Motorsteuerung

// Thermistoren:
volatile int adc_pin[6]; // Analog-Digital-Wandler Pins
volatile int adc_wert[6]; // Wandler Rohwerte
volatile float th_kohm[6]; // Widerstand in kOhm
volatile int i_kohm[6]; // Widerstand in kOhm, da sprintf nicht mit float funktioniert

// Endstops:
volatile int end_pin[6]; // 3 x min und 3 x max
volatile unsigned long ends = 0;

// FETs:
volatile byte fet_pin[6]; // Pins zum Schalten der FETs
volatile int fet; // FET-Steuerung

//Hilfsvariablen:
char zeile1[21]; 
char zeile2[21]; 
char zeile3[21]; 
char zeile4[21]; 
long zeit;
volatile int ii = 0, jj = 0, kk = 0, ll = 0, mm = 0;

// initialize the library with the numbers of the interface pins RS EN DB4 DB5 DB6 DB7
LiquidCrystal lcd(42, 43, 44, 45, 46, 47);

//*******************************************************************************************************************
void lcdWriteByte(uint8_t c, uint8_t rs)
{
    digitalWrite(42, rs);
    digitalWrite(44, c & 0x10);
    digitalWrite(45, c & 0x20);
    digitalWrite(46, c & 0x40);
    digitalWrite(47, c & 0x80);
    delayMicroseconds(1);   // Daten muessen stabil anliegen
    digitalWrite(43, HIGH); 
    delayMicroseconds(20);  // Write data into RAM 38 us (evtl. reichen 20)
    digitalWrite(43, LOW);
    digitalWrite(44, c & 0x01);
    digitalWrite(45, c & 0x02);
    digitalWrite(46, c & 0x04);
    digitalWrite(47, c & 0x08);
    delayMicroseconds(1);   // Daten muessen stabil anliegen
    digitalWrite(43, HIGH); 
    delayMicroseconds(20);  // Write data into RAM 38 us (evtl. reichen 20)
    digitalWrite(43, LOW);
}

//*******************************************************************************************************************
void lcdWriteScreen() {
  //  lcdWriteByte(1, LOW); // wie lcd.clear();
  lcdWriteByte(2, LOW); // Cursor Home;
  delay(2);
  for (mm = 0; mm <= 19; mm++){
    lcdWriteByte(zeile1[mm], HIGH); // RS = HIGH fuer Daten schreiben
  }
  for (mm = 0; mm <= 19; mm++){
    lcdWriteByte(zeile3[mm], HIGH); // RS = HIGH fuer Daten schreiben
  }
  for (mm = 0; mm <= 19; mm++){
    lcdWriteByte(zeile2[mm], HIGH); // RS = HIGH fuer Daten schreiben
  }
  for (mm = 0; mm <= 19; mm++){
    lcdWriteByte(zeile4[mm], HIGH); // RS = HIGH fuer Daten schreiben
  }
}

//*******************************************************************************************************************
void lcdClearScreen() {
  sprintf(zeile1, "                    "); 
  sprintf(zeile2, "                    "); 
  sprintf(zeile3, "                    "); 
  sprintf(zeile4, "                    "); 
}

//*******************************************************************************************************************
void setup() {
  delay(500);
  // LCD: 
  lcd.begin(20, 4); // Spalten, Zeilen

  //  Analog-Digital-Wandler:
  adc_pin[5] = A5; // AUX-ADC
  adc_pin[4] = A4; // Thermistor-ADC
  adc_pin[3] = A3; // Thermistor-ADC
  adc_pin[2] = A2; // Thermistor-ADC
  adc_pin[1] = A1; // Thermistor-ADC
  adc_pin[0] = A0; // Thermistor-ADC
  for (ii = 0; ii <= 5; ii++){
    pinMode(adc_pin[ii], INPUT); 
  }

  // Drehencoder:
  enc_pin[0] = 48; // Encoder E = Knopf
  enc_pin[1] = 50; // Encoder B, im Rastpunkt am Übergang?
  enc_pin[2] = 52; // Encoder A, im Rastpunkt stabil
  for (ii = 0; ii <= 2; ii++){
    pinMode(enc_pin[ii], INPUT_PULLUP);
  }

  // Endstops:
  end_pin[0] = 28; // x_min
  end_pin[1] = 30; // y_min
  end_pin[2] = 32; // z_min
  end_pin[3] = 34; // x_max
  end_pin[4] = 36; // y_max
  end_pin[5] = 38; // z_max
  for (ii = 0; ii <= 5; ii++){
    pinMode(end_pin[ii], INPUT_PULLUP); // Nur für mechanische Endstops!
  }

  // Buzzer:
  pinMode(41, OUTPUT);
  digitalWrite(41, HIGH);
  delay(500);
  digitalWrite(41, LOW);

  // Back-Button:
  pinMode(71, INPUT_PULLUP);

  // FETs:
  fet_pin[0] = 7;  // FET 1 = Heizbett
  fet_pin[1] = 8;  // FET 2
  fet_pin[2] = 9;  // FET 2
  fet_pin[3] = 11; // FET 2
  fet_pin[4] = 12; // FET 2
  fet_pin[5] = 13; // FET 2
  for (ii = 0; ii <= 5; ii++){
    pinMode(fet_pin[ii], OUTPUT); 
    digitalWrite(fet_pin[ii],  LOW);
  }
  fet = 0; // Steuerung der FETs
  p1ms_fet = 10000; // 1 sec. nix machen (nach Start, am Ende von setup)
  
  // Schrittmotoren:
  // X-Achse:
  dir_pin[0] = 23; // Stepper Dir pins: 0=x, 1=y, 2=z, 3=e0, 4=e1, 5=e2
  stp_pin[0] = 24; // Stepper Step pins: x, y, z, e0, e1, e2
  ena_pin[0] = 26; // Stepper enable pins: x, y, z, e0, e1, e2
  // Y-Achse:
  dir_pin[1] = 16; // Stepper Dir pins: 0=x, 1=y, 2=z, 3=e0, 4=e1, 5=e2
  stp_pin[1] = 17; // Stepper Step pins: x, y, z, e0, e1, e2
  ena_pin[1] = 22; // Stepper enable pins: x, y, z, e0, e1, e2
  // Z-Achse:
  dir_pin[2] = 3; // Stepper Dir pins: 0=x, 1=y, 2=z, 3=e0, 4=e1, 5=e2
  stp_pin[2] = 2; // Stepper Step pins: x, y, z, e0, e1, e2
  ena_pin[2] = 15; // Stepper enable pins: x, y, z, e0, e1, e2
  // E1-Achse:
  dir_pin[3] = A6; // Stepper Dir pins: 0=x, 1=y, 2=z, 3=e0, 4=e1, 5=e2
  stp_pin[3] = A7; // Stepper Step pins: x, y, z, e0, e1, e2
  ena_pin[3] = A8; // Stepper enable pins: x, y, z, e0, e1, e2
  // E2-Achse:
  dir_pin[4] = A9; // Stepper Dir pins: 0=x, 1=y, 2=z, 3=e0, 4=e1, 5=e2
  stp_pin[4] = A10; // Stepper Step pins: x, y, z, e0, e1, e2
  ena_pin[4] = A11; // Stepper enable pins: x, y, z, e0, e1, e2
  // E3-Achse:
  dir_pin[5] = 53; // Stepper Dir pins: 0=x, 1=y, 2=z, 3=e0, 4=e1, 5=e2
  stp_pin[5] = 51; // Stepper Step pins: x, y, z, e0, e1, e2
  ena_pin[5] = 49; // Stepper enable pins: x, y, z, e0, e1, e2
  // PINs setzen:
  for (ii = 0; ii <= 5; ii++){
    pinMode(dir_pin[ii], OUTPUT); 
    pinMode(stp_pin[ii], OUTPUT); 
    pinMode(ena_pin[ii], OUTPUT); 
    digitalWrite(dir_pin[ii], LOW);
    digitalWrite(stp_pin[ii], LOW);
    digitalWrite(ena_pin[ii], LOW);
    stp_soll[ii] = 0;
    stp_ist[ii] = 0;
    stp_0[ii] = 0; // Startpunkt (Integrationskonstante)
  }
  // Steuerung der Motoren
  vv = 0.32; // Geschwindigkeit / (stp / 0.1ms)
  motor = 0; 
  p1ms_start = p1ms;
  digitalWrite(ena_pin[motor], HIGH); // enable motor

  // SD-Karte:
  pinMode(14,  INPUT_PULLUP); // SD detect
  for (chipSelect = 4; chipSelect < 11; chipSelect += 6){ // beide Karten testen: 4 = onboard, 10 = LCD
    lcdClearScreen();
    if (chipSelect ==  4) sprintf(zeile1, "onboard Micro-SD:   ");
    if (chipSelect == 10) sprintf(zeile1, "LCD board SD-card:  ");
    lcdWriteScreen();
    pinMode(chipSelect, OUTPUT);
    if (!SD.begin(chipSelect)) { // see if the card is present and can be initialized
      sprintf(zeile2, "failed              ");
      sd_ok = 9;
    } else {
      sprintf(zeile2, "OK                  ");
      sd_ok = 2;
    }
    if (sd_ok == 2) {
      File dataFile = SD.open("SD_test.txt"); // open the file
      if (dataFile) {  // if the file is available, read from it:
        pos = 0;
        while (dataFile.available() && (pos < 20)) {
          zeile3[pos++] = dataFile.read(); 
        }
        dataFile.close();
      } else { // if the file isn't open, pop up an error:
        sprintf(zeile3, "can't open file     ");
      } 
    }
    lcdWriteScreen();
    delay(2000); // Damit SD-Text lesbar ist
  }

  // EEPROM:
  sprintf(zeile1, "Press Back Bottom   "); 
  sprintf(zeile2, "for EEPROM test     "); 
  sprintf(zeile3, "or wait 2 s         "); 
  sprintf(zeile4, "                    "); 
  lcdWriteScreen();

  zeit = millis();
  do {
    if (digitalRead(71) == LOW) {
      Wire.begin(); 
      b1 = 123; // Testbytes
      writeEEPROM(dev_adr, eep_adr, b1);
      b2 = readEEPROM(dev_adr, eep_adr);
      lcdClearScreen();
      sprintf(zeile1, "EEPROM: %3u         ", b2);
      lcdWriteScreen();
      delay(2000); // Damit EEPROM-Text lesbar ist
    }
  } while((millis() - zeit) < 2000);

  // Timer-Interrupt starten:
  Timer3.attachInterrupt(aufruf10kHz);
  Timer3.start(100); // Calls every 0.1ms
}

//*******************************************************************************************************************
void fets_steuern() {
  // FETs Steuern:
  if ((p1ms_fet + 10000) < p1ms) {
    digitalWrite(fet_pin[fet], HIGH);   // FET einschalten
    fet ++; // nächster FET
    if (fet > 5) {
      fet = 0;
    }
    p1ms_fet = p1ms;
    digitalWrite(fet_pin[fet], LOW);   // FET ausschalten    
  }
//  digitalWrite(fet_pin[0], HIGH);   // Heizbett zum Test immer ein
}

//*******************************************************************************************************************
void drehencoder() {
  enc = 0;
  enc = enc + 100 * digitalRead(48); 
  enc = enc + 10 * digitalRead(50); 
  enc = enc + 1 * digitalRead(52); 
}

//*******************************************************************************************************************
void endstops() {
  ends = 0;
  for (ll = 0; ll <= 5; ll++){
    ends = ends + pow(10, ll) * digitalRead(end_pin[ll]);
  }
}

//*******************************************************************************************************************
void lcd_ausgabe() {
  for (mm = 0; mm <= 4; mm++){
    i_kohm[mm] = 0.5 + th_kohm[mm]; // Da sprintf nicht mit float funktioniert
  }
  sprintf(zeile1, "RADDS Test, Thermi.:"); 
  sprintf(zeile2, "%4d%4d%4d%4d%4d", i_kohm[0], i_kohm[1], i_kohm[2], i_kohm[3], i_kohm[4]);  
  sprintf(zeile3, "Encod. %03d ADC5 %4d", enc, adc_wert[5]);  
  sprintf(zeile4, "Endstops: %06d cd%1d", ends, sd_cd);  
  lcdWriteScreen();
}

//*******************************************************************************************************************
void adc_auslesen() {
  for (kk = 0; kk <= 5; kk++){
    adc_wert[kk] = analogRead(adc_pin[kk]); // ersten Wert verwerfen
    adc_wert[kk] = analogRead(adc_pin[kk]);
    th_kohm[kk] = 4.7 * adc_wert[kk] / ((float) (1023.0001 - adc_wert[kk]));
    if (th_kohm[kk] > 998.6) th_kohm[kk] = 998.6;
  }
}

//*******************************************************************************************************************
void buzzer() {
  if (digitalRead(71) == LOW) {
    digitalWrite(41, HIGH);
  } else {
    digitalWrite(41, LOW);
  }
}

//*******************************************************************************************************************
inline void steps_berechnen() {
  // Motoren jeweils eine Umdrehung vor und wieder zurück
  tt = (float) (p1ms - p1ms_start);
  stp_soll[motor] = stp_0[motor] + (int) (vv * tt);

  if (((vv > 0.0) && (stp_soll[motor] > 3200)) || ((vv < 0.0) && (stp_soll[motor] < 0))) { // Vorwärtslauf und Rückwärtslauf steuern
    digitalWrite(ena_pin[motor], LOW); // disable motor
    stp_0[motor] = stp_ist[motor];  // Startwert für nächste Bewegung
    motor ++;  // nächster Motor
    if (motor > 5) {
      motor = 0;  
      vv = -1.0 * vv; // Richtungsumkehr
    }
    digitalWrite(ena_pin[motor], HIGH); // enable motor
    p1ms_start = p1ms;  // Startzeit für aktuellen Motor
  }
}  

//*******************************************************************************************************************
inline void steps_ausfuehren() {
  // Achsen evtl. verfahren:
  for (jj = 0; jj <= 5; jj++){
    if (stp_soll[jj] != stp_ist[jj]) {
      if (stp_soll[jj] > stp_ist[jj]) { // positive Richtung
        digitalWrite(dir_pin[jj], LOW);   // DIR
        stp_ist[jj] ++;
      } else { // negative Richtung
        digitalWrite(dir_pin[jj], HIGH);   // DIR
        stp_ist[jj] --;
      }
      digitalWrite(stp_pin[jj], HIGH);   // Step machen
      delayMicroseconds(1);
      digitalWrite(stp_pin[jj], LOW);   // Step abschliessen
    }
  }
}

//*******************************************************************************************************************
void writeEEPROM(int deviceaddress, unsigned int eeaddress, byte data) {
  Wire.beginTransmission(deviceaddress);
  Wire.write((int)(eeaddress >> 8));   // MSB
  Wire.write((int)(eeaddress & 0xFF)); // LSB
  Wire.write(data);
  Wire.endTransmission();
 
  delay(6); // EEPROM braucht etwas Zeit zum Speichern (max 5 ms)
}
 
//*******************************************************************************************************************
byte readEEPROM(int deviceaddress, unsigned int eeaddress) {
  byte rdata = 0xFF;
 
  Wire.beginTransmission(deviceaddress);
  Wire.write((int)(eeaddress >> 8));   // MSB
  Wire.write((int)(eeaddress & 0xFF)); // LSB
  Wire.endTransmission();
  Wire.requestFrom(deviceaddress, 1);
  if (Wire.available()) {
    rdata = Wire.read();
  }
  return rdata;
}

//*******************************************************************************************************************
void aufruf10kHz(){ // Interrupt für die Motorbewegung
  p1ms++;
  lcd_aus++;
  steps_berechnen();
  steps_ausfuehren();
}

//*******************************************************************************************************************
void loop() {
  adc_auslesen();
  drehencoder();
  fets_steuern();
  endstops();
  buzzer();
  sd_cd = digitalRead(14); 
  if (lcd_aus > 1000) {
    lcd_aus = 0;
    lcd_ausgabe();
  }
}






















